#include "CApp.h"
#include "imgui/imgui.h"
#include "imgui/imgui_impl_opengl3.h"
#include "imgui/imgui_impl_sdl.h"
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_keycode.h>
#include <SDL2/SDL_video.h>
#include <cmath>
#include <cstddef>
#include <iostream>
#include <string>

// logging function
void log(std::string input) { std::cout << input << "\n"; }

// the constructor
CApp::CApp() {
  isRunning = true;
  window = NULL;
  gl_context = NULL;
}

bool CApp::OnInit() {
  if (SDL_Init(SDL_INIT_EVERYTHING) < 0) {
    return false;
  }

  // GL versions
  const char *glsl_version = "#version 150";
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, 0);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);

  // Creat Window
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
  SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);

  SDL_WindowFlags window_flags =
      (SDL_WindowFlags)(SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE |
                        SDL_WINDOW_ALLOW_HIGHDPI);

  window =
      SDL_CreateWindow("Welcome to the Terror Dome", SDL_WINDOWPOS_CENTERED,
                       SDL_WINDOWPOS_CENTERED, 720, 720, window_flags);

  gl_context = SDL_GL_CreateContext(window);
  SDL_GL_MakeCurrent(window, gl_context);

  if (window != NULL) {
    std::cout << "\n> Initialized Window .....\n";
  } else {
    return false;
  }

  // ImGUI stuff??
  IMGUI_CHECKVERSION();
  ImGui::CreateContext();
  ImGuiIO &io = ImGui::GetIO();
  (void)io;
  ImGui::StyleColorsDark();
  ImGui_ImplSDL2_InitForSDLRenderer(window);
  ImGui_ImplOpenGL3_Init(glsl_version);

  std::cout << "> Returning from Init() .....\n";
  return true;
}

int CApp::OnExecute() {
  SDL_Event event;

  if (OnInit() == false) {
    return -1;
  }

  while (isRunning) {
    while (SDL_PollEvent(&event) != 0) {
      OnEvent(&event);
      ImGui_ImplSDL2_ProcessEvent(&event);
    }
    ImGui_ImplOpenGL3_NewFrame();
    ImGui_ImplSDL2_NewFrame();
    ImGui::NewFrame();
    OnLoop();
    OnRender();
  }
  return 0;
}

void CApp::OnEvent(SDL_Event *event) {
  if (event->type == SDL_QUIT) {
    isRunning = false;
  }
}

void CApp::OnLoop() {

  ImGui::Begin("Instrument Window");

  static const ImGuiInputTextFlags flags = ImGuiInputTextFlags_EnterReturnsTrue;

  std::string welcomeMessage =
      "Welcome to the avionics entrance package.\nUse the W/S keys to control "
      "pitch, the A/D keys to control yaw, and the Q/E keys to control roll. "
      "Key-down events are framerate controlled and will appear faster or "
      "slower depending on where the application is being run.";

  ImGui::Text("Hello, Avionics");
  ImGui::TextWrapped(welcomeMessage.c_str());

  static float pitch = 0.0f;
  static float yaw = 0.0f;
  static float roll = 0.0f;

  ImGui::Text("Adjust Pitch");
  if (ImGui::Button("-##1")) {
    pitch -= 1;
    std::cout << "entered: "
              << "pitch decrease"
              << "\n";
  }
  ImGui::SameLine();
  if (ImGui::Button("+##1")) {
    pitch += 1;
    std::cout << "entered: "
              << "pitch increase"
              << "\n";
  }

  ImGui::Text("Adjust Yaw");
  if (ImGui::Button("-##2")) {
    yaw -= 1;
    std::cout << "entered: "
              << "yaw decrease"
              << "\n";
  }
  ImGui::SameLine();
  if (ImGui::Button("+##2")) {
    yaw += 1;
    std::cout << "entered: "
              << "yaw increase"
              << "\n";
  }

  ImGui::Text("Adjust Roll");
  if (ImGui::Button("-##3")) {
    roll -= 1;
    std::cout << "entered: "
              << "roll decrease"
              << "\n";
  }
  ImGui::SameLine();
  if (ImGui::Button("+##3")) {
    roll += 1;
    std::cout << "entered: "
              << "roll increase"
              << "\n";
  }

  if (ImGui::IsKeyDown(26)) {
    pitch -= 1;
  }
  if (ImGui::IsKeyDown(22)) {
    pitch += 1;
  }
  if (ImGui::IsKeyDown(4)) {
    yaw -= 1;
  }
  if (ImGui::IsKeyDown(7)) {
    yaw += 1;
  }
  if (ImGui::IsKeyDown(20)) {
    roll -= 1;
  }
  if (ImGui::IsKeyDown(8)) {
    roll += 1;
  }

  ImGui::Text("Pitch at: %3.2f degrees", pitch);
  ImGui::Text("Yaw at: %3.2f degrees", yaw);
  ImGui::Text("Roll at: %3.2f degrees", roll);

  ImGui::SliderFloat("Pitch", &pitch, -180.0f, 180.0f, "%.2f");
  ImGui::SliderFloat("Yaw", &yaw, -180.0f, 180.0f, "%.2f");
  ImGui::SliderFloat("Roll", &roll, -180.0f, 180.0f, "%.2f");

  if (ImGui::TreeNode("Keyboard & Navigation State")) {
    ImGui::Text("Keys down:");
    for (int i = 0; i < IM_ARRAYSIZE(io->KeysDown); i++)
      if (ImGui::IsKeyDown(i)) {
        ImGui::SameLine();
        ImGui::Text("%d (0x%X) (%.02f secs)", i, i, io->KeysDownDuration[i]);
      }
    ImGui::Text("Keys pressed:");
    for (int i = 0; i < IM_ARRAYSIZE(io->KeysDown); i++)
      if (ImGui::IsKeyPressed(i)) {
        ImGui::SameLine();
        ImGui::Text("%d (0x%X)", i, i);
      }
    ImGui::Text("Keys release:");
    for (int i = 0; i < IM_ARRAYSIZE(io->KeysDown); i++)
      if (ImGui::IsKeyReleased(i)) {
        ImGui::SameLine();
        ImGui::Text("%d (0x%X)", i, i);
      }
    ImGui::Text("Keys mods: %s%s%s%s", io->KeyCtrl ? "CTRL " : "",
                io->KeyShift ? "SHIFT " : "", io->KeyAlt ? "ALT " : "",
                io->KeySuper ? "SUPER " : "");
    ImGui::Text("Chars queue:");
    for (int i = 0; i < io->InputQueueCharacters.Size; i++) {
      ImWchar c = io->InputQueueCharacters[i];
      ImGui::SameLine();
      ImGui::Text("\'%c\' (0x%04X)", (c > ' ' && c <= 255) ? (char)c : '?', c);
    } // FIXME: We should convert 'c' to UTF-8 here but the functions are not
      // public.

    ImGui::Text("NavInputs down:");
    for (int i = 0; i < IM_ARRAYSIZE(io->NavInputs); i++)
      if (io->NavInputs[i] > 0.0f) {
        ImGui::SameLine();
        ImGui::Text("[%d] %.2f (%.02f secs)", i, io->NavInputs[i],
                    io->NavInputsDownDuration[i]);
      }
    ImGui::Text("NavInputs pressed:");
    for (int i = 0; i < IM_ARRAYSIZE(io->NavInputs); i++)
      if (io->NavInputsDownDuration[i] == 0.0f) {
        ImGui::SameLine();
        ImGui::Text("[%d]", i);
      }

    ImGui::Button("Hovering me sets the\nkeyboard capture flag");
    if (ImGui::IsItemHovered())
      ImGui::CaptureKeyboardFromApp(true);
    ImGui::SameLine();
    ImGui::Button("Holding me clears the\nthe keyboard capture flag");
    if (ImGui::IsItemActive())
      ImGui::CaptureKeyboardFromApp(false);
    ImGui::TreePop();
  }

  // ImGui::ShowDemoWindow();

  ImGui::End();
}

void CApp::OnRender() {
  ImGui::Render();
  glViewport(0, 0, 1280, 720);
  glClearColor(0.45f, 0.55f, 0.60f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT);
  ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
  SDL_GL_SwapWindow(window);
}

void CApp::OnExit() {
  std::cout << "> Exiting...";
  ImGui_ImplOpenGL3_Shutdown();
  ImGui_ImplSDL2_Shutdown();
  ImGui::DestroyContext();

  SDL_GL_DeleteContext(gl_context);
  gl_context = NULL;
  SDL_DestroyWindow(window);
  window = NULL;
  SDL_Quit();
}
